@import url("reset-to-defaults.css");

#window {
  background-image: none;
  background-color: lime;
  border-radius: 0;
  border-style: none;
}

#nowindow {
  background-color: red;
}

#window#nowindow {
  background-color: blue;
}
